static char rcsid[] = "@(#)$Id: date.c,v 2.1 2018/07/08 18:01:01 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.1 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *           or  Kari Hurtta <elm@elmme-mailer.org>
 ******************************************************************************
 *  Based on Elm 2.4 src/date.c. That code was following copyright:
 *
 *  The Elm Mail System 
 *
 * 			Copyright (c) 1988-1992 USENET Community Trust
 * 			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************
 *  Incorparated Elm 2.5 code from src/date.c
 *  That code have following copyright:
 *
 *  The Elm Mail System  
 *
 *                      Copyright (c) 1988-1995 USENET Community Trust
 * 			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

#include "def_addr.h"

DEBUG_VAR(Debug,__FILE__,"addr");

#include <limits.h>

#define MONTHS_IN_YEAR	11	/* 0-11 equals 12 months! */
#define FEB		 1	/* 0 = January 		  */
#define DAYS_IN_LEAP_FEB 29	/* leap year only 	  */

#define leapyear(year)  (((year) % 4 == 0) && (((year) % 100 != 0) || ((year) % 400 == 0)) )

static int  days_in_month[] = { 31,    28,    31,    30,    31,     30, 
				31,     31,    30,   31,    30,     31,  -1};

struct tm * days_ahead_tm(days)
     int days;
{
    /** return the date (Day, Mon Day, Year) of the date
	'days' days after today.  
    **/

    time_t      junk =                  /* time in seconds....		     */
	time(NULL);                     /* this must be here for it to work! */

    struct tm * the_time = 	        /* Time structure, see CTIME(3C)     */
	localtime(&junk);
    
    if (the_time) {

	DPRINT(Debug,16,(&Debug,
			 "days_ahead_tm: localtime gives tm_mday %d (1-31), tm_mon %d (0-11), tm_year %d (Year - 1900), tm_wday %d (0-6, Sunday = 0)\n",
			 the_time->tm_mday,
			 the_time->tm_mon,
			 the_time->tm_year,
			 the_time->tm_wday
			 ));

	/* If expires-have-time is shown then also time is shown */

	DPRINT(Debug,16,(&Debug,
			 "days_ahead_tm: localtime gives tm_sec %d (0-60), tm_min %d (0-59), tm_hour %d (0-23)\n",
			 the_time->tm_sec,
			 the_time->tm_min,
			 the_time->tm_hour));

	if (the_time->tm_hour < 17) {
	    DPRINT(Debug,16,(&Debug,
			     "days_ahead_tm: Rounding expires time to 18:00:00\n"));

	    the_time->tm_sec  = 0;
	    the_time->tm_min  = 0;
	    the_time->tm_hour = 18;
	} else {
	    DPRINT(Debug,16,(&Debug,
			     "days_ahead_tm: Rounding expires time to 23:59:59\n"));

	    the_time->tm_sec  = 59;
	    the_time->tm_min  = 59;
	    the_time->tm_hour = 23;
	}
		
	/* increment the day of the week */	
	if (INT_MAX - days < the_time->tm_wday) {
	    DPRINT(Debug,16,(&Debug,
			     "days_ahead_tm: Overflow tm_wday, days=%d\n",
			     days));

	    return NULL;	    
	}       
	the_time->tm_wday = (the_time->tm_wday + days) % 7;
	
	
	/* increment the day of the month... */
	if (INT_MAX - days <  the_time->tm_mday) {
	    DPRINT(Debug,16,(&Debug,
			     "days_ahead_tm: Overflow tm_mday, days=%d\n",
			     days));
	    
	    return NULL;	    
	}		
	the_time->tm_mday += days;

	/* adjust overflowing day of the month ... */
	
	while (the_time->tm_mday > days_in_month[the_time->tm_mon]) {
	    
	    if (the_time->tm_mon == FEB && leapyear(the_time->tm_year+1900)) {
		if (the_time->tm_mday > DAYS_IN_LEAP_FEB) {
		    the_time->tm_mday -= DAYS_IN_LEAP_FEB;
		    the_time->tm_mon += 1;
		} else
		    break;            /* Is Feb 29, so leave */
	    } else {
		the_time->tm_mday -= days_in_month[the_time->tm_mon];
		the_time->tm_mon += 1;
	    }

	    /* check the month of the year */
	    if (the_time->tm_mon > MONTHS_IN_YEAR) {

		if (the_time->tm_year < INT_MAX) {
		    the_time->tm_mon -= (MONTHS_IN_YEAR + 1);
		    the_time->tm_year += 1;
		} else {
		    DPRINT(Debug,16,(&Debug,
				     "days_ahead_tm: Overflow tm_year, days=%d\n",
				     days));
		    return NULL;
		}
	    }
	}
	
	DPRINT(Debug,16,(&Debug,
			 "days_ahead_tm: %d days gives tm_mday %d (1-31), tm_mon %d (0-11), tm_year %d (Year - 1900), tm_wday %d (0-6, Sunday = 0)\n",
			 days,
			 the_time->tm_mday,
			 the_time->tm_mon,
			 the_time->tm_year,
			 the_time->tm_wday));

	DPRINT(Debug,16,(&Debug,
			 "days_ahead_tm: tm_sec %d (0-60), tm_min %d (0-59), tm_hour %d (0-23)\n",
			 the_time->tm_sec,
			 the_time->tm_min,
			 the_time->tm_hour));
	

    } else {
	DPRINT(Debug,1,(&Debug, 
			"days_ahead_tm: %d days -- localtime failed, time=%ld\n",
			days,
			(long)junk
			));
    }

    return the_time;
}

/* Return 1 if succeed */
int days_ahead_hdrval(ptr,buffer,size,print_time)
     struct expanded_expires *ptr;
     char            * buffer;
     int               size;
     int               print_time;
{
    const struct tm * expire_time =
	get_expanded_expires_value(ptr);
    
    if (expire_time) {

	if (expire_time->tm_wday < 0 ||
	    expire_time->tm_wday >=
	    sizeof (arpa_dayname) / sizeof (arpa_dayname[0]) ||
	    expire_time->tm_mday < 1 ||
	    expire_time->tm_mday > 31 ||
	    expire_time->tm_mon < 0 ||
	    expire_time->tm_mon >=
	    sizeof (arpa_monname) / sizeof (arpa_monname[0])) {
	    
	    DPRINT(Debug,16,(&Debug,
			     "days_ahead_hdrval=0: date failure tm_mday %d (1-31), tm_mon %d (0-11), tm_year %d (Year - 1900), tm_wday %d (0-6, Sunday = 0)\n",
			     expire_time->tm_mday,
			     expire_time->tm_mon,
			     expire_time->tm_year,
			     expire_time->tm_wday));
	    
	    return 0;
	}
	
	if (print_time) {
	    int tz_mins = 0;
	    const char * zone = NULL;
	    char sign = '+';
	    
	    if (expire_time->tm_sec < 0  ||
		expire_time->tm_sec > 60 ||
		expire_time->tm_min < 0  ||
		expire_time->tm_min > 59 ||
		expire_time->tm_hour < 0 ||
		expire_time->tm_hour > 23) {
		
		DPRINT(Debug,16,(&Debug,
				 "days_ahead_hdrval: time failure tm_sec %d (0-60), tm_min %d (0-59), tm_hour %d (0-23)\n",
				 expire_time->tm_sec,
				 expire_time->tm_min,
				 expire_time->tm_hour));
		
		
		goto fail_time;
	    }

#ifdef TZNAME_USE_TM_ZONE
	    tz_mins = expire_time->tm_gmtoff / 60;
	    zone    = expire_time->tm_zone;

	    DPRINT(Debug,16,(&Debug,
			     "days_ahead_hdrval: tm_gmtoff %d, tm_zone %s\n",
			     expire_time->tm_gmtoff,
			     expire_time->tm_zone ? expire_time->tm_zone : "NULL"));	    
#endif
	    
	    if (!zone) {
		time_t cached_time =
		    get_expanded_expires_cached_time(ptr);
		
		DPRINT(Debug,16,(&Debug,
				 "days_ahead_hdrval: Assuming local time zone, cached_time=%ld\n",
				 cached_time));

		if (cached_time > 0) {
		    tz_mins = get_tz_mins(cached_time);
		} else {
		    DPRINT(Debug,16,(&Debug,
				     "days_ahead_hdrval:  timezone not available\n"));
		    goto fail_time;
		}
		
		zone = get_tz_name(expire_time);

	    }
	    
	    if (tz_mins < 0) {
		sign = '-';
		tz_mins = -tz_mins;
	    }

	    if (zone &&
		( '-' == zone[0] ||
		  '+' == zone[0])) {
		DPRINT(Debug,16,(&Debug,
				"days_ahead_hdrval: Ignoring zone name %s\n",
				zone));
		zone = NULL;
	    }
	    
	    DPRINT(Debug,16,(&Debug,
			     "days_ahead_hdrval: tz_mins %d sign %c zone %s\n",
			     tz_mins,sign,zone ? zone : "NULL"));
	    
	    elm_sfprintf(buffer, size,
			 FRM("%s, %d %s %d %02d:%02d:%02d %c%02d%02d (%s)"),
			 arpa_dayname[expire_time->tm_wday],
			 expire_time->tm_mday,
			 arpa_monname[expire_time->tm_mon],
			 1900+expire_time->tm_year,
			 expire_time->tm_hour,
			 expire_time->tm_min,
			 expire_time->tm_sec,
			 sign,
			 tz_mins / 60,
			 tz_mins % 60,
			 zone ? zone : "");

	} else {
	fail_time:
	    elm_sfprintf(buffer, size,
			 FRM("%s, %d %s %d"),
			 arpa_dayname[expire_time->tm_wday],
			 expire_time->tm_mday,
			 arpa_monname[expire_time->tm_mon],
			 1900+(expire_time->tm_year));
	}

	DPRINT(Debug,16,(&Debug,
			 "days_ahead_hdrval=1: %s\n",
			 buffer));
	
	return 1;
    }

     DPRINT(Debug,16,(&Debug,
		     "days_ahead_hdrval=0: No expire_time\n"));

    return 0;
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */

